// WialonSDK file
// import 'dart:io';
import 'dart:convert';
import 'package:http/http.dart' as http;
import './exceptions.dart';

/// Wialon SDK for dart
class WialonSdk {
  /// boolean to activate the development debugger if set to true
  bool _isDevelopment = false;

  /// Wialon Session Id
  String _sessionId = '';

  /// Default parameters that will be added to all calls to the Wialon API
  Map<String, dynamic> _defaultParams = {};

  /// Wialon user Id
  int _userId = null;

  /// Base url for the request
  String _baseUrl = '';

  /// Wialon url
  String _url = 'hst-api.wialon.com';

  /// constructor
  WialonSdk({
    isDevelopment = false,
    sessionId = '',
    String scheme = 'https',
    String host = 'hst-api.wialon.com',
    int port = 0,
    Map<String, dynamic> extraParams = const {}
  }) {
    this._isDevelopment = isDevelopment;
    this._sessionId = sessionId;

    if (port < 0) {
      throw SdkException(message: 'Invalid port, must be greater than 0');
    } else if (port > 0) {
      this._baseUrl = "${scheme}://${host}:${port}";
    } else {
      this._baseUrl = "https://hst-api.wialon.com";
    }

    this._url = host;

    this._baseUrl += "/wialon/ajax.html?";

    this._defaultParams = {
      ...extraParams
    };
  }

  /// Set sesion Id
  void setSessionId({ String sessionId }) {
    this._sessionId = sessionId;
  }

  /// get Wialon Session Id
  String getSessionId() {
    return this._sessionId;
  }

  /// Get Wialon user Id
  int getUserId() {
    return this._userId;
  }

  /// Login into Wialon with the API token
  Future<Map<String, dynamic>> login({ String token }) async {
    dynamic result =  await this.call(method: 'token/login', parameters: {
      'token': token
    });

    this._userId = result['user']['id'];
    this._sessionId = result['eid'];

    return result;
  }

  /// Logout of Wialon
  Future<void> logout() async {
    await this.call(method: 'core/logout', parameters: {'action': 'logout'});

    this._userId = 0;
    this._sessionId = '';
  }

  /// Rever geocoding, returns name of the location that corresponds to the latitude and longitude given
  Future<String> reverseGeocoding({
    /// latitude
    double latitude,
    /// longitude
    double longitude,
    /// flags
    int flags = 1255211008
  }) async {
    final Map<String, double> coordinates = {
      'lon': longitude,
      'lat': latitude
    };

    final String url = "https://geocode-maps.wialon.com/${this._url}/gis_geocode?coords=[${jsonEncode(coordinates)}]&flags=${flags}&uid=${this._userId}";

    return await http.post(url).then((response) {
      List result = null;
      try {
        result = jsonDecode(response.body);
      } catch (e) {
        throw SdkException(message: "Internal error: ${e}");
      }
      return result[0].toString();
    }).catchError((error) => {
      throw SdkException(message: "Internal error: ${error}")
    });
  }

  /// Make a call to the Wialon API
  Future<Map<String, dynamic>> call({ String method, dynamic parameters }) async {
    dynamic arguments = {
      ...this._defaultParams
    };

    if (parameters is List) {
      arguments = [...parameters];
    } else {
      arguments.addAll(parameters);
    }

    Map<String, dynamic> payload = {
      'svc': method,
      'params': jsonEncode(arguments),
      'sid': this._sessionId
    };

    if (this._isDevelopment) {
      this._debugPrinter(
        title: "SDK Call: ${method}",
        content: "Payload: ${payload}\n" +
                "URL: ${this._baseUrl}svc=${payload['svc']}" +
                "&params=${payload['params']}&sid=${payload['sid']}"
      );
    }

    Map<String, dynamic> result = await http.post(this._baseUrl, body: payload).then((http.Response response) {
      Map<String, dynamic> result;
      try {
        result = jsonDecode(response.body);
      } catch (e) {
        throw SdkException(message: "Internal error: ${e}");
      }

      if (result['error'] == null) {
        return result;
      }else if (result['error'] == 0) {
        return result;
      } else {
        throw WialonError(
          code: result['error'],
          details: result['reason'] == null ? '' : result['reason']
        );
      }
    }).catchError((error) => {
      throw SdkException(message: "Internal error: ${error}")
    });

    return result;
  }

  /// debug printer
  void _debugPrinter({ String title, dynamic content }) {
    print("*******" * 10);
    print("${title}\n${content}");
    print("*******" * 10);
  }
}